# by the saved changes moeten ook de watchfolder gegevens worden bijgevoegd als dict
# ombouwen naar mongodb
#-----------------------------------------------------------------------
import os
import time


#-----------------------------------------------------------------------
if "." in __name__:
	from modules import config
	from modules import tasks
	from modules import database
else:
	import config
	import tasks
	import database


#-----------------------------------------------------------------------
#
# add entry to database
#
#-----------------------------------------------------------------------
def create_collection():
	if "watchfolders" not in database.list_collections():
		watchfolders					= database.Blueprint()
		watchfolders.name 				= database.String()
		watchfolders.name.default		= "watchfolder"
		watchfolders.root 				= database.Directory()
		watchfolders.project			= database.String()
		watchfolders.project.default	= "default"
		watchfolders.on_new				= database.String()
		watchfolders.on_change			= database.String()
		watchfolders.on_deleted			= database.String()
		watchfolders.arguments			= database.Dictionary()
		watchfolders.extensions			= database.List()
		watchfolders.active				= database.Boolean()
		watchfolders.status.default		= True
		
		coll 							= database.Collection( "watchfolders" )
		coll.blueprint 					= watchfolders
		coll.create()
		
		

#-----------------------------------------------------------------------
def print_change(**kwargs):
	print( "#"*16)
	print("Change found in folder: %s" % kwargs["root"])
	print("%s: %s" % (kwargs["change"].upper(), kwargs["path"]))
	print( "#"*16)


#-----------------------------------------------------------------------
class WatchFolder(object):
	def __init__( self, **kwargs ):
		self.auto_dispatch 	= True
		self.name 			= None
		self.root 			= None
		self.extensions 	= []
		self.recursive 		= False

		self.on_new			= print_change
		self.on_changed 	= print_change
		self.on_deleted 	= print_change

		for key in [ "auto_dispatch", "name", "root", "extensions", "recursive" ]:
			if key in kwargs.keys():
				setattr( self, key, kwargs[key] )

		if not self.exists:
			raise ValueError( "Folder %s doesn't exist" % self.root )

		columns = { "root":database.TEXT(), "path": database.TEXT(), "modtime":database.TEXT(), "change":database.TEXT() }
		#name=os.path.path.basename(self.root)[1]
		sqlitename="_".join(database.PATH().path_to_list(self.root)[1:])
		self.FILE=database.File(os.path.join(config.LOCAL, "%s.sqlite"%sqlitename))

		for item in [ "SAVED", "TMP" ]:
			if item not in self.FILE:
				table=database.Table( item, columns=columns )
				self.FILE.append(table)

		columns={"path":database.TEXT(), "watchfolder": database.TEXT(), "change":database.TEXT()}
		if "changes" not in self.FILE:
			table = database.Table("changes", columns=columns)
			self.FILE.append(table)

		self.CHANGES = self.FILE["changes"]

	@property
	def exists(self):
		try:
			return os.path.isdir(self.root)
		except:
			return False

	def check_for_changes( self ):
		if self.exists:
			self.FILE["TMP"].clear()

			for item in list_folder( self.root, extensions = self.extensions ):
				if item[ "path" ] != self.FILE.uri:
					#print "Adding to TMP:", item
					self.FILE["TMP"].append( **item )

			for row in self.FILE.compare( self.FILE["TMP"], self.FILE["SAVED"], "path" ):
				rowdict = { "change" : "new", "path" : row["path"], "watchfolder" : self.name }
				self.FILE["SAVED"].append( row )
				self.CHANGES.append( **rowdict )

			for row in self.FILE.compare( self.FILE["SAVED"], self.FILE["TMP"], "path" ):
				rowdict = { "change" : "deleted", "path" : row["path"], "watchfolder" : self.name }
				row.delete()
				self.CHANGES.append( **rowdict )

			for row in self.FILE.compare( self.FILE["TMP"], self.FILE["SAVED"], "modtime", "path" ):
				tmp_row = self.FILE["SAVED"].find(path=row["path"])
				if tmp_row !=None:
					tmp_row["modtime"]=row["modtime"]

					rowdict = { "change" : "changed", "path" : row[ "path" ], "watchfolder" : self.name }
					self.CHANGES.append( **rowdict )

			if self.auto_dispatch:
				self.dispatch_changes()

		else:
			raise ValueError("Watchfolder %s doesn't exist" % self.root)

	#-----------------------------------------------------------------------
	def dispatch_changes(self):
		for row in self.CHANGES:
			error=None
			if row["change"]!= None:
				# error moet in een error log komen en niet de change database!
				try:
					trigger=getattr( self, "on_%s" % row["change"] )
					trigger(**row.as_dict())
				except:
					error="Error executing: something"
				row.delete()



#-----------------------------------------------------------------------
def list_folder( root, recursive = False, extensions = [] ):
	for item in os.listdir( root ):
		path = None
		if extensions:
			if os.path.splitext( item )[1] in extensions:
				path = item
		else:
			path = item

		if path:
			row = {}
			fullpath = os.path.join( root, path )
			if os.path.exists( fullpath ):
				modtime = os.path.getmtime( fullpath )
				row = { "root" : root, "path" : fullpath, "modtime" : modtime }
				#print row
				yield row


#-----------------------------------------------------------------------
def start_watching():
	# addon van maken
	pass
	# ~ while True:

		# ~ for row in gestion.WATCHFOLDERS.findall(watch=True):
			# ~ print( "Going to watch folder: ", row["root"])
			# ~ wf=watchfolder.WatchFolder(**row.as_dict())
			# ~ wf.auto_dispatch=False
			# ~ if wf.exists:
				# ~ wf.check_for_changes()
			# ~ else:
				# ~ row["watch"] = False
		# ~ time.sleep(5)


#-----------------------------------------------------------------------
def dispatch_changes():
	# addon van maken
	pass
	# ~ while True:
		# ~ for row in gestion.WATCHFOLDERS.findall(watch=True):
			# ~ wf=watchfolder.WatchFolder(**row.as_dict())
			# ~ if wf.exists:
				# ~ try:
					# ~ wf.on_new=addons.LOADED[row["on_new"]]
					# ~ wf.on_deleted=addons.LOADED[row["on_deleted"]]
					# ~ wf.on_changed=addons.LOADED[row["on_changed"]]
				# ~ except ValueError as e:
					# ~ print( "Error in getting addon: %s" % e)
				# ~ wf.dispatch_changes()
			# ~ else:
				# ~ row["watch"]=False
		# ~ time.sleep(5)







import os
import sys

# ~ from modules import config
# ~ from modules import database
# ~ from modules import web
from modules import tasks
from modules import cli
# ~ from modules import node


#-----------------------------------------------------------------------
__all__ = []


#-----------------------------------------------------------------------
for item in os.listdir( os.path.split( __file__ )[0] ):
	if item[0] not in ["_", "."]:
		file_name, extension = os.path.splitext( item )
		if extension == ".py":
			__all__.append( file_name )
			

#-----------------------------------------------------------------------
from . import *
LOADED = None


#-----------------------------------------------------------------------
def available_addons():
	return [ item for item in LOADED.keys() ]

	
def list_addons():
	print( "#" * 32 )
	print( "Available addon(s):" )
	for item in available_addons():
		print( item )
	print( "#" * 32 )


cli.add_command( list_addons )


#-----------------------------------------------------------------------
def load():
	loaded = {}
	for item in __all__:
		imp = getattr( sys.modules[ __name__ ],  item )
		for item in dir( imp ):
			obj = getattr( imp, item )
			try:
				if isinstance( obj, tasks.Addon ):
					loaded[ obj.func_name ] = obj
			except Exception as e:
				tasks.write_to_log( str( e ) )
	
	setattr(  sys.modules[ __name__ ],  "LOADED", loaded )


#-----------------------------------------------------------------------
def get_addon( name ):
	if name in available_addons():
		return LOADED[ name ]
	else:
		tasks.write_to_log( "No addon found with name: %s" % name )


load()



import os
import sys
import time
import shutil
import argparse
from importlib import reload as reload


# -----------------------------------------------------------------------
WORKING_DIR = os.path.split( __file__ )[0]
sys.path.append( WORKING_DIR )


# -----------------------------------------------------------------------
import modules
from modules import config
from modules import cli

if not config.file_exists:
	config.create_file()
	config.load_from_file()


print( f"Initializing database" )
from modules import web
web.init( __name__ )

print( "Initializing gestion." )
from modules import database
database.init()

# ~ print( 1 )
from modules import node
# ~ print( 2 )
from modules import tasks
# ~ print( 3 )
from modules import queue
# ~ print( 4 )
from modules import watchfolder
# ~ print( 5 )


#-----------------------------------------------------------------------
def main():
	if config.FIRST_RUN:
		print( f"First run, creating gestion config folder: {config.GESTION_FOLDER}" )
		# ~ default = os.path.join( config.GESTION_FOLDER, "default" )
		if not os.path.exists( config.GESTION_FOLDER ):
			os.makedirs( config.GESTION_FOLDER )
		
		for folder in [ "static", "templates", "default" ]:
			src = os.path.join( WORKING_DIR, "template", folder )
			dst = os.path.join( config.GESTION_FOLDER, folder )
			print( f"Copying: {src} -> {dst}" )
			shutil.copytree( src, dst, dirs_exist_ok = True )
		
		config.FIRST_RUN = False


	if config.PROJECT not in node.available_projects():
		print( "Creating default project, not found in available projects." )
		node.create_project( config.PROJECT, config.PROJECT_ROOT )

	if len( sys.argv ) > 1:
		cli.parse_args_and_execute( sys.argv )
	else:
		start()


# -----------------------------------------------------------------------
def start():
	if not database.database_is_running():
		print( f"Error connecting to database on ip {config.DATABASE_IP}:{config.DATABASE_PORT}" )
		print( "Going to quit gestion" )
		exit()
	
	if config.PROJECT not in node.available_projects():
		print( "creating project , not found in VlVBLE PROJSECT" )
		node.create_project( config.PROJECT, config.PROJECT_ROOT )
	
	node.set_project( config.PROJECT )

	#-------------------------------------------------------------------
	pages			 	= []
	processes_to_start 	= []

	for name in [ name for name in dir( modules ) if not name.startswith("__") ]:
		mod = getattr( modules, name )
		for item in dir( mod ):
			obj = getattr( mod, item )
			# ~ if isinstance( obj, type ):
			if isinstance( obj, web.Page ):
				pages.append( obj )
			elif isinstance( obj, tasks.Task ):
				processes_to_start.append( obj )
				# ~ web.create_route( obj )


	#-------------------------------------------------------------------
	import addons
	print( "Loading addons" )
	addons_path = os.path.join( os.path.split( __file__ )[0], "addons"  )
	for f in os.listdir( addons_path ):
		if f[0] not in ["_", "."]:
			addon_name, extension = os.path.splitext( f )
			if extension.lower() == ".py":
				addon = getattr( addons, addon_name )
				for item in dir( addon ):
					obj = getattr( addon, item )
					# ~ if isinstance( obj, type ):
					if isinstance( obj, web.Page ):
						for route in obj:
							obj = f"/addon/{route}"
						pages.append( obj )
					elif isinstance( obj, tasks.Task ):
						processes_to_start.append( obj )


	print( "Creating pages" )
	for page in pages:
		web.create_route( page )


	print( "Starting tasks" )
	for proc in processes_to_start:
		print( "Starting task: %s" % proc.func_name )
		proc.execute()

	print( f"Connected to database on ip {config.DATABASE_IP}:{config.DATABASE_PORT}" )
	web.start_server()



#-----------------------------------------------------------------------
def set_config( **kwargs ):
	for key in kwargs.keys():
		setattr( config, key.upper(), kwargs[key] )

	config.save_to_file()

cli.add_command( set_config )


# -----------------------------------------------------------------------
if __name__ == "__main__":
	main()







#-----------------------------------------------------------------------
import os
import sys
import json
import time
import shutil
from pathlib import Path, PurePath
from configparser import ConfigParser


#-----------------------------------------------------------------------
if "." in __name__:
	from modules import config
	from modules import database
	from modules import web
	from modules import cli
	from modules import tasks

else:
	import config
	import database
	import web
	import cli
	import tasks


#-----------------------------------------------------------------------
# ~ def create_collections():
	# ~ [ "mounts", "software", "projects", "log", "users" ]
projects 				= database.Blueprint()
projects["name"]		= database.String()
projects["root"]		= database.Path()
projects.set_local( True )
projects_collection 	= database.create_collection( "projects", projects )


software 				= database.Blueprint()
software["name"]		= database.String()
software["path"]		= database.Path()
software["extension"]	= database.String() # misschien list met strings?
software.set_local( True )
software_collection 	= database.create_collection( "software", software )


# https://stackoverflow.com/questions/54750273/pymongo-and-ttl-wrong-expiration-time
nodes 				= database.Blueprint()
nodes["host"]		= database.Host()
nodes["status"]		= database.String()
nodes["heartbeat"]	= database.Integer()
nodes_collection 	= database.create_collection( "nodes", nodes )


log 				= database.Blueprint()
log["type"]			= database.String()
log["log"] 			= database.String()
log["userhost"]		= database.Userhost()
log["date"]			= database.Date()
log.set_local( False )

log_collection		= database.create_collection( "log", log )



#-----------------------------------------------------------------------
#
# node tasks
#
#-----------------------------------------------------------------------
# https://stackoverflow.com/questions/54750273/pymongo-and-ttl-wrong-expiration-time
def heartbeat():
	while True:
		doc = nodes_collection.find_one( host = config.HOST )
		time.sleep( 10 )

heartbeat_task 		= tasks.Task( heartbeat )
heartbeat_task.wait = False


#-----------------------------------------------------------------------
#
# commonly used functions included
#
#-----------------------------------------------------------------------
def map_path( path ):
	pass


#-----------------------------------------------------------------------
def add_software( name, path, extension = None ):
	coll 				= database.get_collection( "software" )
	doc 				= coll.create_document()
	doc["name"]			= name
	doc["path"]			= path
	doc["extension"]	= extension
	coll.append( doc )

cli.add_command( add_software )


#-----------------------------------------------------------------------
def available_software():
	coll 	= database.get_collection( "software" )
	results = coll.find()
	
	print( f"Available software on {config.HOST}" )
	for item in results:
		print( f'{item["name"]}' )
	return [ item["name"] for item in results ]

cli.add_command( available_software )


#-----------------------------------------------------------------------
def get_software( name = None, extension = None ):
	path = None
	if name in available_software():

		kwargs = {}
		if name != None:
			kwargs[ "name" ] = name
		if extension != None:
			 kwargs[ "extension" ] = extension
		# print( kwargs )
		document = database.Collection("software").find_one( **kwargs )
		# print( document )
		if document:
			path = document[ "path" ]
		
		
	
	if path is None:
		raise ValueError( f"Cannot find specified software on {config.HOST}" )
	else:
		print( path )
		return path

cli.add_command( get_software )


#-----------------------------------------------------------------------
# ~ https://linuxtut.com/en/9f4dd46c96a86608f519/
# ~ parser.optionxform = str
#-----------------------------------------------------------------------
def create_workspace( name, root, **kwargs ):
	
	configparser 			 = ConfigParser()
	configparser[ "global" ] = kwargs
	workspace_file 			 = os.path.join( root, f"{name.lower()}.gestion" )
	print( f"Creating workspace with file:{workspace_file}" )
	with open( workspace_file, 'w' ) as f:
		configparser.write( f )


def load_workspace( workspace_file ):
	root, base 		= os.path.split( workspace_file )
	workspace_file 	= os.path.join( root, base.lower() )
	print( f"Load workspace from file:{workspace_file}" )
	configparser 	= ConfigParser()
	configparser.read( workspace_file )
	return configparser[ "global" ]


def workspace_exists( name ):
	doc = projects_collection.find_one( name = name )
	if doc == None:
		return False
	else:
		return file_exists( os.path.join( doc[ "root" ], f"{name}.gestion" ) )


#-----------------------------------------------------------------------
def create_project( name, root, **kwargs ):
	result 	= projects_collection.find_one( name = name )
	
	if result == None:
		if folder_exists( root ):
			if is_file( root ):
				raise ValueError( f'"{root}" is a file, select a directory instead.' )
			elif not is_folder( root ):
				raise ValueError( f'{root} does not exist.' )
			
			workspace_file = os.path.join( root, f"{name.lower()}.gestion" )
			if file_exists( workspace_file ):
				# ~ load_workspace( workspace_file )
				print( f'A project has already been created for {root}, not creating workspace.' )
			
			else:
				create_workspace( name, root, **kwargs )

			doc = projects_collection.create_document()
			doc["name"] 	= name
			doc["root" ] 	= root
			projects_collection.append( doc )
		# kwargs in een bestandje in de root van het project
	
	if workspace_exists( name ):
		set_project( name )


cli.add_command( create_project )


#-----------------------------------------------------------------------
def set_project( name ):
	doc = projects_collection.find_one( name = name )
	if doc != None:
		if workspace_exists( name ):
			workspace_file 		= os.path.join( doc[ "root" ], f"{name}.gestion" )
			config.WORKSPACE	= load_workspace( workspace_file )
			config.PROJECT_ROOT	= doc[ "root" ]
			config.PROJECT_NAME	= name
		else:
			raise Exception( f"Cannot find a workspace file in {doc[ 'root' ]}" )

	else:
		raise Exception( f"No project found in projects database with the name: {name}" )
	
	

cli.add_command( set_project )


#-----------------------------------------------------------------------
def available_projects():
	results = projects_collection.find()
	print("#"*32)
	print( "Projects available:" )
	for result in results:
		print( "%s: %s" % ( result["name"],  result["root"] ) )
	print("#"*32)
	return [ item["name"] for item in results ]

cli.add_command( available_projects )


#-----------------------------------------------------------------------
def delete_project( name ):
	doc = projects_collection.find_one( name = name )
	if doc != None:
		projects_collection.delete_one( doc._id )


cli.add_command( delete_project )


#-----------------------------------------------------------------------
def write_to_log( to_log:str ):
	coll 			= database.get_collection( "log" )
	doc		 		= coll.create_document()
	doc[ "log" ]	= to_log
	return coll.append( doc )



#-----------------------------------------------------------------------
#
# os helper functions
#
#-----------------------------------------------------------------------
def is_file( path ):
	path = Path( path )
	if path.exists():
		return path.is_file()
	else:
		return False


def is_folder( path ):
	path = Path( path )
	if path.exists():
		return path.is_dir()
	else:
		return False


def file_exists( path ):
	if Path( path ).exists():
		return is_file( path )
	else:
		return False


def folder_exists( path ):
	if Path( path ).exists():
		return is_folder( path )
	else:
		return False


def create_folder( path ):
	path = Path( path )
	if not path.exists():
		path.mkdir( parents=True )
	else:
		raise FileExistsError( f"Can not create directory {path}, it already exists.")


def copy( src, dst ):
	if is_folder( src ):
		shutil.copytree( src, dst, dirs_exist_ok = True )
	elif is_file( src ):
		shutil.copy( src, dst )

#-----------------------------------------------------------------------
#
# web pages for the node
#
#-----------------------------------------------------------------------
index_page				= web.Page( "/" )
settings_page			= web.Page( "settings", template = web.CollectionTemplate( software_collection ) )
project_page			= web.Page( "projects", template = web.CollectionTemplate( projects_collection ) )

log_template 	= web.CollectionTemplate( log_collection )
log_page 		= web.Page( "log", template = log_template )


#-----------------------------------------------------------------------
def toggle_status():
	config.STATUS = not config.STATUS

toggle_by_click 			= web.SimpleClick( "toggle_status" )
toggle_by_click.function 	= toggle_status







import os, json, zipfile

#-----------------------------------------------------------------------
class Layer( dict ):
	def __init__(self, layer, parent=None):
		dict.__init__( self )
		self.parent = parent

		for key in layer.keys():
			self[key] = layer[key]

#-----------------------------------------------------------------------#
class File( object ):
	def __init__( self, path, as_json = True ):
		self.as_json 	= as_json
		self.root 		= {}
		self.path 		= path
		self.tmp_path 	= os.path.split( path )[0]
		self.extension 	= os.path.splitext( path )[1]

		try:
			with zipfile.ZipFile( path, "r" ) as z:
				z.extractall( self.tmp_path )
		except:
			raise ValueError("Cannot extract moho file: %s" % path)

		if as_json:
			self.JSON_to_DICT()
			self.layers 		= self.get_layers()
			self.grouplayers 	=	 self.get_grouplayers()
		else:
			self.as_string 		= self.open_file()

	#-----------------------------------------------------------------------#
	def zipfile_info(self):
		info = {}

		contents 	= [ "Project.%sproj" % self.extension.replace( ".",  "" ), "preview.jpg" ]
		z 			= zipfile.ZipFile( self.path )
		try:
			for item in contents:
				info[ item ] = z.getinfo( item )
		except:
			pass
		return info

	#-----------------------------------------------------------------------#
	def find(self, key):
		for key in self.root.keys():
			print(key)

	#-----------------------------------------------------------------------#
	def pretty_print(self):
		return json.dumps(self.root, indent=4, sort_keys=True)

	#-----------------------------------------------------------------------#
	def open_file(self):
		with open(os.path.join(self.tmp_path, "Project.%sproj" % self.extension.replace(".",  ""))) as f:
			contents = f.read()
		return contents

	#-----------------------------------------------------------------------#
	def replace_in_file(self, _from, _to):
		newstring = self.as_string.replace(_from, _to)
		self.root = json.loads(newstring.decode('utf-8'))

	#-----------------------------------------------------------------------#
	def JSON_to_DICT(self):
		self.root = json.loads( self.open_file() )

	#-----------------------------------------------------------------------#
	def get_layers(self):
		return return_layers(self.root)

	def get_styles(self):
		return self.root["styles"]

	def get_grouplayers(self):
		return return_layergroups(self.root)

	def get_imagelayers(self, recursive=True):
		return return_imagelayers(self.root, recursive=recursive)

	def get_file_info(self):
		return self.decoded.keys()

	@property
	def project_data(self):
		return self.root["project_data"]
	
	@property
	def start_frame(self):
		return self.root["project_data"]["start_frame"]
	
	@start_frame.setter
	def start_frame(self, frame):
		self.root["project_data"]["start_frame"]=frame
	
	@property
	def end_frame(self):
		return self.root["project_data"]["end_frame"]
		
	@end_frame.setter
	def end_frame(self, frame):
		self.root["project_data"]["end_frame"]=frame

	@property
	def fps(self):
		return self.root["project_data"]["fps"]
		
	@fps.setter
	def fps(self, fps):
		self.root["project_data"]["fps"]=fps

	def write_changes(self):
		with open(os.path.join(self.tmp_path, "Project.%sproj" % self.extension.replace(".",  "")), "w") as f:
			f.write(json.dumps(self.root))

	def save(self, path):
		newextension = os.path.splitext(path)[1]
		print("Project.%sproj" % newextension.replace(".",  ""))
		print(os.path.join(self.tmp_path, "Project.%sproj" % self.extension.replace(".",  "")))

		self.write_changes()

		with zipfile.ZipFile(path, "w", zipfile.ZIP_DEFLATED) as f:
			f.write(os.path.join(self.tmp_path, "Project.%sproj" % self.extension.replace(".",  "")), "Project.%sproj" % newextension.replace(".",  ""))
			f.write(os.path.join(self.tmp_path,"preview.jpg"), "preview.jpg")

	def replace_psd(self):
		pass

#-----------------------------------------------------------------------#
def return_layers(root, recursive=True):
	if not recursive:
		return iter_layers(root)
	else:
		return iter_layers(root, return_type="imagelayer")

def return_layergroups(root):
	return [layer for layer in return_layers(root) if layer["type"].lower()=="grouplayer"]

def return_imagelayers(root, recursive=True):
	return [layer for layer in return_layers(root, recursive=recursive) if layer["type"].lower()=="imagelayer"]

#-----------------------------------------------------------------------#
def versioned_file(path):
	_path, _extension = os.path.splitext(path)
	_path, _file = os.path.split(_path)
	current_version=_file.split("_")[-2]
	_author = _file.split("_")[-1]
	_file="_".join(_file.split("_")[:-2])

	string_list=[]
	int_list=[]

	string_len=len(current_version)
	for i in range(string_len):
		try:
			int(current_version[i])
			int_list.append(current_version[i])

		except:
			string_list.append(current_version[i])

	new_version=int(current_version[len(string_list):])+1
	vString="".join(string_list)+("%0"+str(len(int_list))+"d")%new_version
	
	newfile= os.path.join(_path, (_file+"_"+vString+"_"+_author+_extension))
	return newfile


#-----------------------------------------------------------------------#
def as_string(layer):
	psdkeys=[]

	psdkeys=[key for key in layer.keys() if "psd" in key]
	ss="%s: %s:" + "\n%s "*len(psdkeys)
	allkeys=["name",  "image_path"]+psdkeys

	return ss % (tuple(layer[key] for key in allkeys))


#-----------------------------------------------------------------------#
def iter_layers(root, return_type="layers", layers=[], reset=False ):
	if reset:
		layers=[]

	if hasattr(root, "root"):
		root = root.root

	for layer in root["layers"]:
		if layer["type"].lower()==return_type:
			layers.append(layer)

		elif "layers" in layer.keys():
			iter_layers(layer, return_type=return_type, layers=layers)

	return layers


#-----------------------------------------------------------------------#
def return_psd_layers(root):
	layers=[]
	for layer in iter_layers(root, return_type="imagelayer", reset=True):
		if "psd_layer" in layer.keys():
			layers.append(layer)

	return layers


#-----------------------------------------------------------------------#
def rgb_to_float(rgb, as_string=False):
	new_r=round( (rgb[0]/255.0), 6)
	new_g=round( (rgb[1]/255.0), 6)
	new_b=round( (rgb[2]/255.0), 6)
	if as_string:
		return '{"r":%s,"g":%s,"b":%s,"a":1.0}' % (new_r, new_g, new_b)
	else:
		return {"a":1.0, "r":new_r, "g":new_g, "b":new_b}


#-----------------------------------------------------------------------#
def get_last_version(path):
	mod_time={}
	for _f in os.listdir(path):
		if os.path.splitext(os.path.join(path, _f))[1] == ".anime":
			mod_time[os.path.getctime(os.path.join(path, _f))]=os.path.join(path, _f)

	try:
		return mod_time[sorted(mod_time.keys())[-1]]
	except:
		return None




import subprocess

from modules import config
from modules import node
from modules import tasks

#-----------------------------------------------------------------------
from . import moho

#-----------------------------------------------------------------------
#
# Switches for renderling moho
#
#-----------------------------------------------------------------------
"""
Moho -r CS10_001_Dinosaur_v020_L.moho -o Renders/CS10_001_Dinosaur/CS10_001_Dinosaur.png -f png -start 1 -end 2800 -halfsize no -multithread no -extrasmooth yes
Moho -r CS10_001_Dinosaur_v020_L.moho -o Renders/CS10_001_Dinosaur/CS10_001_Dinosaur.png -f png -start 1 -end 2800 -halfsize no -multithread no -extrasmooth yes -layercomp AllComps -createfolderforlayercomp yes -addlayercompsuffix yes
"""

#-----------------------------------------------------------------------
def render( file, start = 1, end = 25, multithread = True, halfsize = False, f = "png", extrasmooth = True, layercomp = False ):
	software 	= node.get_software( "moho" )
	choice 		= { False : "no", True : "yes" }
	output 		= "/nog/te/bepalen"


	if software != None:
		cli = 			[ 	software,
							"-r", 			file,
							"-o", 			output,
							"-f", 			f,
							"-start",		start,
							"-end",			end,
							"-halfsize", 	choice[ halfsize ],
							"-multithread",	choice[ multithread ],
							"-extrasmooth", choice[ extrasmooth ],
						]
			
		if bool( layercomp ) == True:

			cli.extend( [ 	"-layercomp", "AllComp",
							"-createfolderforlayercomp", choice[ True ],
							"-addlayercompsuffix", choice[ True ],
						] )
						
		proc = subprocess.Popen( cli )
		proc.wait()
	
	else:
		raise ValueError( "Cannot find path to moho executable" )
		

#-----------------------------------------------------------------------
#
# check if moho software is available, if so than create addon
#
#-----------------------------------------------------------------------
if "moho" in node.available_software():
	mohorender_addon = tasks.Addon( render )
	mohorender_addon.wait = True

# ~ mohorender_addon = tasks.Addon( render )
# ~ mohorender_addon.wait = True




#-----------------------------------------------------------------------
import os
import sys
import time
import inspect
import builtins

from flask import Flask
from flask import Response
from flask import stream_with_context
from flask import redirect
from flask import render_template
from flask import request
from flask import url_for
from flask import json
from markupsafe import Markup


#-----------------------------------------------------------------------
routes_to_be_created 	= []
server 					= None


#-----------------------------------------------------------------------
if "." in __name__:
	from modules import config
	from modules import database
else:
	import config
	import database


#-----------------------------------------------------------------------
def init( name ):
	setattr( sys.modules[ __name__ ], "server", Flask( name, static_folder = config.STATIC_FOLDER, template_folder = config.TEMPLATE_FOLDER ) ) # static folder default in home static


#-----------------------------------------------------------------------
def start_server( debug = True, use_reloader = False ):
	threaded = False
	server.run( debug = debug, host = config.WEB_IP, port = config.WEB_PORT, threaded = threaded, use_reloader = use_reloader )


#-----------------------------------------------------------------------
#
#	template classes
#
#-----------------------------------------------------------------------
class Template():
	def __init__( self, html_template ):
		self.html_template	= html_template
		self.header			= []


	@property
	def html_template( self ):
		return self._html_file


	@html_template.setter
	def html_template( self, name ):
		self._html_file = f"{name}.html"


	def render_page( self, *args, **kwargs ):
		return render_template( self.html_template, header = self.header, *args, **kwargs )


#-----------------------------------------------------------------------
class DefaultTemplate( Template ):
	def __init__( self ):
		Template. __init__( self, "default" )


#-----------------------------------------------------------------------
class CollectionTemplate( Template ):
	if "." in __name__:
		from modules import database
	else:
		import database


	def __init__( self, collection:database.Collection ):
		Template. __init__( self, "default" )
		self.collection = collection
		self.header		= collection.blueprint.keys( ignore=[ "id", "hidden", "saved" ] )


	def return_page( self, *args, **kwargs ):
		# aanpassen naar return_page ?
		return render_template( self.html_template, header = self.header, **kwargs )


#-----------------------------------------------------------------------
class Redirect( Template ):
	def __init__( self ):
		self.url = None


	def return_page( self ):
		return url_for( "/%s" % self.url )


#-----------------------------------------------------------------------
class Referrer( Template ):
	def __init__( self ):
		pass


	def return_page( self ):
		return redirect(  request.referrer )


#-----------------------------------------------------------------------
def collection_function():
	print( request.referrer )


#-----------------------------------------------------------------------
class Page( list ):
	def __init__( self, route:str, template = DefaultTemplate(), methods = [ "GET" ] ):
		list.__init__( self, [ route ] )
		# ~ self.module		= None
		self.methods	= methods
		self.template	= template
 
 
	def __iter__( self ):
		# replace '.' with '_" or '/' or delete?
		for item in list.__iter__( self ):
			if item.startswith("/"):
				yield item
			else:
				yield f"/{item}"


	@property
	def name( self ):
		return list.__getitem__(self, 0 )


	@property
	def template( self ):
		return self._template


	@template.setter
	def template( self, template ):
		if issubclass( template.__class__, Template ):
			self._template = template
		else:
			raise Exception( f"Error in {self.__class__.__name__}: Need a class subclassed from a Template, not a {template.__class__.__name__}" ) 


	@property
	def methods( self ):
		return self._methods


	@methods.setter
	def methods( self, methods:list = [], *args ):
		if not hasattr( self, "_methods" ):
			self._methods = []

		methods.extend( args )
		for method in methods:
			if method not in self._methods:
				self._methods.append( method )
				
	
	@property
	def function( self ):
		# function expcets a function, but this isn ot the case, so need anther name?
		if hasattr( self, "_function" ):
			return self._function
		else:
			return None


	@function.setter
	def function( self, func ):
		if callable( func ):
			self._func_parser( func )
			self._function = func
		else:
			raise ValueError( f"Need a function, not a {type(func)}" )


	def _func_parser( self, func ):
		pass


	def wrapper_function( self, *args, **kwargs ):
		start 	= time.time()
		if self.function:
			results	= self.function( *args, **kwargs )
		else:
			if issubclass( self.template.__class__, CollectionTemplate ):
				results = self.collection_wrapper_function( *args, **kwargs )
			else:
				results = f"Rendering page: {self.name}"
		end 	= time.time()

		return self.template.render_page( data = results, page = self, config = config )
	
	#---------------------------------------------------------------
	#
	# speciale class maken voor collection
	# met deze function, voor het overzicht
	#
	#---------------------------------------------------------------
	def collection_wrapper_function( self, *args, **kwargs ):
		# perhaps check for collection_function
		referrer 	= request.referrer
		root 		= request.url_root
		args 		= request.args

		if "sort" in args:
			key 	= args.get( "sort" )
			sort	= -1

			if referrer != None:
				old_args = referrer[len(root):]
				if key in old_args:
					if self.template.collection._sort[1] == -1:
						sort = 1
			self.template.collection._sort = ( key, sort )

		return self.template.collection.find_in_cache( **kwargs )


#-----------------------------------------------------------------------
class SimpleClick( Page ):
	def __init__( self, route, *args, **kwargs ):
		Page.__init__( self, route, *args, **kwargs )
		self.needs_document		= False
		self.needs_collection	= False
		self.is_filter			= False
		# ~ self.needs_id			= False


	def _func_parser( self, func ):
		has_document = False
		sig = inspect.signature( func)
		for parameter in sig.parameters:
			if issubclass( sig.parameters[parameter].annotation, database.Document ):
				self.needs_document = True
				break


	def wrapper_function( self, *args, **kwargs ):
		# ~ from modules import database
		# ~ if 
		# ~ print( "id in  request.args", ("id" in  request.args) )
		if self.needs_document:
			# ~ print( "needs a document" )
			if "id" in request.args:
				_id 	= request.args.get( "id" )

			if  "db" in request.args:
				_db 	= request.args.get( "db" )

			if _id	!= None and _db != None:
				coll 	= database.get_collection( _db )
				doc 	= coll[ database.ObjectId( _id ) ] 
				results = self.function( doc )


		elif self.needs_collection:
			_db 	= request.args.get( "db" )
		# ~ elif _db != None:
			results = self.function( _db, *args, **kwargs )
		
		else:
			results = self.function( *args, **kwargs )
		
		return redirect( request.referrer )


#-----------------------------------------------------------------------\
#
# helper functions
#
#-----------------------------------------------------------------------
def create_route( page ):
	
	if server == None:
		raise Exception( "The server has not yet been initialized!" )
	else:
		# mss de naam veranderen in create_route, page is meer met een template.
		if issubclass( page.__class__, Page ):
			try:
				func_name = f"{page.name}_function"
				for route in page:
					# ~ if issubclass( page.__class__, SimpleClick ):
						# ~ if page.needs_colection:
							# ~ route = f"{route}?db={}" 
					server.add_url_rule( route, func_name, page.wrapper_function, methods = page.methods )
					server.view_functions[ route ] = page.wrapper_function
					print( "Created route(s) for:\t%s " % route )

			except Exception as e:
				print( "Error creating route(s) for: %s " % page.name )
				print( str( e ) )




import os
import sys
import inspect
from socket import gethostname
from getpass import getuser
from configparser import ConfigParser

#-----------------------------------------------------------------------
#
# config based functions and periphelia
#
#-----------------------------------------------------------------------
GESTION_FOLDER_NAME = "gestion_dev"
GESTION_DB_NAME		= "gestion_dev_db000124"
GESTION_WEB_PORT	= 5001
DEFAULT_STATUS		= ["online", "local", "offline" ]	 # voor later gebruik
DEFAULT_STATUS		= False
NODE_TYPE			= [ "client", "master" ] 			# voor later gebruikt
NODE_TYPE			= "client"


#-----------------------------------------------------------------------
defaults = 	{
				"_user" 			: getuser(),
				"_host" 			: gethostname().split( "." )[ 0 ],
				"_home" 			: os.path.expanduser( "~" ),
				"_gestion_folder"	: os.path.join( os.path.expanduser( "~" ), GESTION_FOLDER_NAME ),
				"_template_folder"	: os.path.join( os.path.expanduser( "~" ), GESTION_FOLDER_NAME, "templates" ),
				"_static_folder"	: os.path.join( os.path.expanduser( "~" ), GESTION_FOLDER_NAME, "static" ),
				"_config_file"		: os.path.join( os.path.expanduser( "~" ), GESTION_FOLDER_NAME, "config.ini" ),
				"database_ip" 		: "127.0.0.1", 
				"database_port"		: 27017,
				"web_ip" 			: "127.0.0.1",
				"web_port" 			: GESTION_WEB_PORT,
				"node_type"			: NODE_TYPE,
				"project"			: "default",
				"project_root"		: os.path.join(os.path.expanduser( "~" ), GESTION_FOLDER_NAME, "default" ),
				"_workspace"		: {},
				"gestion_db"		: GESTION_DB_NAME,
				"status"			: DEFAULT_STATUS,
				"_first_run"		: True
			}


#-----------------------------------------------------------------------
class Config( dict ):
	def __init__( self ):
		dict.__init__( self, defaults )
		self.FILE = self.pop( "_config_file" )

		if self.file_exists:
			self.load_from_file()
		
		
	def __getattr__( self, key ):
		env 	= None
		
		if "_" + key.lower() in self.keys():
			env = "_" + key.lower()
		elif key.lower() in self.keys():
			env = key.lower()
		
		if env:
			inst 		= defaults[ env ]
			value		= self[ env ]
			if type( inst ) == type( value ):
				return value
			else: 
				if type( inst ) == bool:
					if self[ env ].lower() == "false":
						attr_as_type = False
					else:
						attr_as_type = True
				else:
					attr_as_type = type( inst )( self[ env ] )
				return attr_as_type
		else:
			return self.__dict__[ key ]


	def __setattr__( self, key, value ):
		env 	= None
		
		if key.lower() in self.keys():
			env = key.lower()
		
		elif "_" + key.lower() in self.keys():
			env = "_" + key.lower()

		if env:
			self[ env ] = value
		else:
			dict.__setattr__( self, key, value )


	def __setitem__( self, key, value ):
		dict.__setitem__( self, key, value )
		if self.file_exists:
			self.save_to_file()
		
		
	def __getitem__( self, key ):
		if self.file_exists:
			self.load_from_file()
		return dict.__getitem__( self,  key )
		
	
	@property
	def file_exists( self ):
		return os.path.exists( self.FILE )
		

	# ~ @property
	# ~ def is_first_run( self ):
		# ~ return self.first_run


	def create_file( self ):
		if not os.path.exists( os.path.split( self.FILE )[ 0 ] ):
			os.makedirs( os.path.split( self.FILE )[ 0 ] )
		self.save_to_file()
		
		
	def save_to_file( self ):
		kwargs = {}
		for key in [ key for key in dict.keys( self ) if not key.startswith("_") ]:
			kwargs[ key ] = dict.__getitem__( self, key )

		configparser 				= ConfigParser()
		configparser[ "global" ] 	= kwargs
		
		with open( self.FILE, 'w' ) as f:
			configparser.write( f )

		
	def load_from_file( self ):
		configparser = ConfigParser()
		configparser.read( self.FILE )
		self.update( **configparser[ "global" ] )
		# ~ print( self )
		# ~ print(dict.__getitem__( self, "status" ) )
		# ~ print( type(defaults["status"])( dict.__getitem__( self, "status" ) ) )


	def reset( self, hard_reset = True ):
		pass


#-----------------------------------------------------------------------
config = Config()
sys.modules[__name__] = config









import 	os
import 	time
import 	psutil
import 	multiprocessing
from 	multiprocessing import Process
from 	multiprocessing import Queue


#-----------------------------------------------------------------------
if "." in __name__:
	from modules import config
	from modules import database
	from modules import web
else:
	import config
	import database
	import web


#-----------------------------------------------------------------------
class Results( object ):
	def __init__( self ):
		self.output = None
		self.error 	= None


#-----------------------------------------------------------------------
class Proc():
	def __init__( self, func ):
		self.func 	 = func
		self.wait	 = False
		self.daemon	 = False
		self.results = None


	#-------------------------------------------------------------------
	def execute( self, *args, **kwargs ):
		if "." in __name__:
			from modules import node
	
		else:
			import node

		node.write_to_log( "Starting proces: %s %s" % ( self.func.__name__ , str( kwargs ) ))
		self.queue		= Queue()
		self.queue.put( { "args":args, "kwargs":kwargs } )
		self.proc 		= Process( target = self.wrapped_func )
		self.proc.name	= self.func.__name__
		
		self.proc.daemon = self.daemon
		self.proc.start()

		if self.wait:
			self.proc.join()
		
	
	def kill( self ):
		proc = psutil.Process( self.pid )
		children = proc.children( recursive=True )
		for child in children:
			child.kill()

		self.proc.terminate()
		if not self.wait:
			self.proc.join()
		self.queue.put( { "output":f"Process killed by user {config.USER}", "error":None } )


	#-------------------------------------------------------------------
	def wrapped_func( self ):
		output 	  = None
		error 	  = None
		arguments = self.queue.get()
		try:
			output = self.func( *arguments[ "args" ], **arguments[ "kwargs" ] )
		except Exception as exc:
			# ~ if f"Process killed by user {config.USER}" in str(exc):
				# ~ output = str(exc)
			# ~ else:
			error = str(exc)
		self.queue.put( { "output":output, "error":error } )



	#-------------------------------------------------------------------
	@property
	def output( self ):
		if hasattr( self, "queue" ):
			self.results = self.queue.get()
			delattr( self, "queue" )

		if self.results != None:
			return self.results[ "output" ]
		else:
			return None

			
	@property
	def error( self ):
		print( "error" )
		if hasattr( self, "queue" ):
			self.results = self.queue.get()
			delattr( self, "queue" )

		if self.results != None:
			return self.results[ "error" ]
		else:
			return None

	
	@property
	def func_name( self ):
		return self.func.__name__


	@property
	def pid(self):
		return self.proc.pid


	@property
	def state(self):
		return self.proc.is_alive()



	@property
	def exitcode( self ):
		return self.proc.exitcode


	@property
	def type(self):
		return self.__class__.__name__.lower()



#-----------------------------------------------------------------------
class Addon( Proc ):
	def __init__( self, func ):
		Proc.__init__( self, func )
		self.daemon = True
		self.wait 	= False


class Task( Proc ):
	def __init__( self, func ):
		Proc.__init__( self, func )
		self.wait	= False


#-----------------------------------------------------------------------
#
# taskmanager-a-like functions
#
#-----------------------------------------------------------------------
def find_process_by_name( name ):
	results = [ proc.info for proc in psutil.process_iter( ["name", "pid"] ) if name.lower() in proc.info["name"].lower() ]
	return results


#-----------------------------------------------------------------------
def find_process_by_port( port:int ):
	processes = [ psutil.Process(pid=proc.pid) for proc in psutil.net_connections() if proc.laddr.port == port ]
	return processes


#-----------------------------------------------------------------------
def manager():
	while True:
		print("Checking for zombified processes")




#-----------------------------------------------------------------------
# ~ def start():
	# ~ pass


#-----------------------------------------------------------------------



import time
from modules import web
from modules import database


def toggle_status( doc:database.Document ):
	print( doc )
	print( "djkf" )


click 			= web.SimpleClick( "test" )
click.function 	= toggle_status



import subprocess


from modules import node
from modules import tasks


def open_file( file ):
	app = node.get_software( name = "filebrowser" )
	if app != None:
		cli = app
		print( cli, file )
		proc = subprocess.Popen( [ cli, file ] )
		proc.wait()
	else:
		raise ValueError( f"Cannot find {app}. If it is installed, add it to the available software" ) 


imageviewer_addon = tasks.Addon( open_file )
imageviewer_addon.wait = False



import os
import sys
import json
import inspect
import pymongo
import pathlib
from bson.objectid 	import ObjectId
from datetime 		import datetime
from markupsafe 	import Markup


#-----------------------------------------------------------------------
#
#	add a watch function, if change happend to collection, add it to an collection_exists
#	check is userhost has already checked if change happened
#	if not already checked and change took place, return true and add userhost to collection changed
#	if collection changed remove all user hosts 
#
#-----------------------------------------------------------------------

#-----------------------------------------------------------------------
#
#	mongod version  v4.4.4
#	pymongo 3.12.0
# 	https://github.com/AlexandreMahdhaoui/MongoNow
#	misschien al extra module installeren zodat er geen echte mongodb hoeft te runnen in eerste instantie
#	of dat je kunt kiezen dat je een connectie wil maken met een server of dat het alleen lokaal runt
#
#-----------------------------------------------------------------------
if "." in __name__:
	from modules import config
else:
	import config


#-----------------------------------------------------------------------
#
#	document, blueprint class
#
#-----------------------------------------------------------------------
class Blueprint( dict ):
	root_id = ObjectId( b"%012d" % 0 )
	def __init__( self ):
		dict.__init__( self )
		if self.__class__ == Blueprint:
			dict.__setitem__( self, "_id", Blueprint.root_id )

		self.set_local( False )


	@property
	def _id( self ):
		if "_id" in dict.keys( self ):
			return dict.__getitem__( self, "_id" )
		else:
			return None


	def is_local( self ):
		return isinstance( self[ "_local" ], Userhost )


	def set_local( self, local:bool ):
		if local == False:
			datatype = Null()
		else:
			datatype = Userhost()

		datatype.hidden = True
		dict.__setitem__( self, "_local", datatype )


	def __setitem__( self, attr, datatype ):
		if attr not in  [ "_id" ] :
			if isinstance( datatype, Datatype ):
				if not attr.startswith( "_" ):
					dict.__setitem__( self, attr, datatype )
				else:
					raise Exception( f"{attr} is an invalid attribute name, it can not start with an underscore!" )
			else:
				raise Exception( f"Can't add {type( datatype )}, need a {database.Datatype}." )
		else:
			raise Exception( f"Can not set '_id' of {self.__class__.__name__}" )


	def __getitem__( self, attr ):
		item = dict.__getitem__( self, attr )
		if attr not in [ "_id" ]:
			if not isinstance( item, Datatype ):
				instance = getattr( sys.modules[ item[ "_module" ] ], item[ "_class" ] )()
				instance.update( item )
				self.__setitem__( attr, instance )
				return instance
			else:
				return item
		else:
			return item


	def keys( self, ignore = None ):
		#---------------------------------------------------------------
		#
		# ignore types are properties like locked and hidden, not the name of keys
		#
		#---------------------------------------------------------------
		if ignore == None:
			ignore = []
		
		keys_to_ignore 			= [] 
		attributes_to_ignore 	= []
		
		for item in ignore:
			if item in [ "saved", "id" ]:
				keys_to_ignore.append( f"_{item}" )
			elif item in [ "locked", "hidden", "unique" ]:
				attributes_to_ignore.append( f"_{item}" )
			else:
				raise ValueError( "Ignorelist can only contain 'id', 'saved', 'unique', 'locked' or 'unique'" )


		keys = list( dict.keys( self ) )
		for item in self.items():
			if item[0] not in [ "_id", "_saved" ]:
				for attr_to_ignore in attributes_to_ignore:
					if item[ 1 ][ attr_to_ignore ] == True:
						result = item[ 1 ][ attr_to_ignore ]
						if item[ 0 ] in keys:
							
							keys.pop( keys.index( item[ 0 ] ) )
			
			elif item[0] in keys_to_ignore:
				if item[ 0 ] in keys:
					keys.pop( keys.index( item[ 0 ] ) )

		return keys


	def set_all_values( self ):
		for key in self.keys( ignore = [ "id", "saved" ] ):
			inst = dict.__getitem__( self, key )
			inst.value_setter()


	def print( self, pretty_print = False ):
		print( self )


#-----------------------------------------------------------------------
class Document( Blueprint ):
	def __init__( self ):
		Blueprint.__init__( self )


	@classmethod
	def create_from_blueprint( cls, blueprint:Blueprint ):
		inst = cls()
		inst.update( blueprint )
		return inst


	def create_new_id( self ):
		dict.__setitem__( self, "_id", ObjectId() )


	def __setitem__( self, attr, value ):
		if attr not in ["_id", "_saved" ]:
			Blueprint.__getitem__( self, attr ).value = value
		else:
			dict.__setitem__( self, attr, value )


	def __getitem__( self, attr ):
#		need to add a reserved keys attribute
		if attr not in [ "_id", "_saved" ]:
			# print( "not in list")
			return Blueprint.__getitem__( self, attr ).value
		else:
			# print( "attr starts with an underscore")
			return dict.__getitem__( self, attr )


	def update( self, kwargs ):
		if "_id" in kwargs:
			if kwargs["_id"] == Blueprint.root_id:
				kwargs.pop( "_id" )
		dict.update( self, kwargs )


	def set_all_values( self, for_insert = True ):
		Blueprint.set_all_values( self )
		if for_insert:
			# ~ # print( "for_insert" )
			_saved = { "_functions":{} }
			for key in self.keys( ignore = [ "id", "saved" ] ):
				datatype				= Blueprint.__getitem__( self, key )
				_saved[ key ]			= datatype.as_html( self )
				_saved[ "_collection" ] = datatype[ "_collection" ]
				_function 				= datatype[ "_function" ]
				if _function != None:
					_saved[ "_functions" ][ key ] = f'{_function[ "_href" ]}'
			_saved[ "_id" ] = self._id
			# ~ print( f"_saved:  {_saved}" )
			dict.__setitem__( self, "_saved", _saved )


#-----------------------------------------------------------------------
#
# Collection class, connection to a database
#
#-----------------------------------------------------------------------
class Collection( object ):
	def __init__( self, name ):
		self._name 		= name
		self._blueprint = None
		self._sort 		= ( "_id", -1 )
		self._limit		= 25


	def __str__( self ):
		return "<Instance of Collection: %s>" % self.name


	def __repr__( self ):
		return "<Instance of Collection: %s>" % self.name


	@classmethod
	def create_from_blueprint( cls, name:str, blueprint:Blueprint ):
		coll = cls( name )
		coll.blueprint = blueprint
		if not coll.exists: 
			coll.database[ name ].insert_one( coll._blueprint )

		return coll


	@property
	def name( self ):
		return self._name


	@name.setter
	def name( self, name ):
		self._name = name


	@property
	def database_name( self ):
		return self.database.name


	@property
	def database( self ):
		return connect_to_database()


	@property
	def exists( self ):
		if self.database_name in list_databases():
			if collection_exists( self.name ):
				result = self.database[ self.name ].find_one( { "_id": Blueprint.root_id } )
				if result != None:
					# ~ self._blueprint = Blueprint.create_from_dict( result )
					return True
				else:
					return False
		else:
			return False


	@property
	def blueprint( self ):
		if not self.exists:
			raise Exception( f"No collection found for {self.name} in database {self.database_name}!" ) 

		else:
			if self._blueprint is None:
				result = self.database[ self.name ].find_one( { "_id": Blueprint.root_id } )
				object_id = 	result.pop( "_id" , None )
				if object_id == Blueprint.root_id:
					blueprint = Blueprint()
					for key in result.keys():
						data 			= result[key]
						instance		= getattr( sys.modules[ data[ "_module" ] ], data[ "_class" ] )()
						instance.update( data )
						if not key.startswith( "_" ):
							blueprint[ key ]= instance
							# ~ if instance.function != None:
								# ~ create_route_for_click( instance, check_exists = True )
						elif key == "_local":
							blueprint.set_local( isinstance( instance, Userhost ) )
					self._blueprint = blueprint

				else:
					raise Exception( f"Need results from collection {self.name} with a root_id, result has a different _id!" )
			return self._blueprint


	@blueprint.setter
	def blueprint( self, blueprint ):
		if isinstance( blueprint, Blueprint ):
			for key in blueprint.keys():
				if key != "_id":
					attr = blueprint[ key ][ "_collection" ] = self.name
			blueprint.set_all_values()
			self._blueprint = blueprint

		else:
			raise Exception( "Need a Blueprint instance, not a %s " % type( blueprint ) )

		# ~ else:
			# ~ raise Exception( "Collection %s already exists!" % self.name )


	def is_local( self ):
		return self.blueprint.is_local()


	def has_changed( self ):
		return check_for_changes( self.name )


	def create_document( self, kwargs:dict = {} ):
		blueprint = self.blueprint
		if self.is_local():
			# ~ blueprint["local"]
			pass
		document = Document.create_from_blueprint( self.blueprint )
		if self.is_local():
			document.local = Userhost()

		for key in kwargs:
			document[ key ] = kwargs[ key ]
		return document


	def create( self ):
		if not self.exists:
			if self._blueprint != None:
				self.database[ self.name ].insert_one( self._blueprint )


	def append( self, document:Document ):
		document.create_new_id()
		document.set_all_values()
		result = self.database[ self.name ].insert_one( document )
		# ~ print( dir( result ) )
		register_database_change( self.name )
		return result.inserted_id


	def extend( self, *documents ):
		self.database[ self.name ].insert_many( *[ doc.set_all_values() for doc in documents ] )
		register_database_change( self.name )


	def clear( self ):
		self.database[ self.name ].delete_many( {} )
		register_database_change( self.name )

	
	def delete_one( self, key ):
		self.database[ self.name ].delete_one( { "_id" : ObjectId(key) } )
		register_database_change( self.name )


	def drop( self ):
		self.database[ self.name ].drop()
		register_database_change( self.name )


	def __getitem__( self, key ):
		# ~ if isinstance( key, slice ):
			# ~ start 	= int( key.start ) + 1
			# ~ end 	= int( key.stop ) + 1
			# ~ results = []
			# ~ for item in list( self.database[ self.name ].find( {} )[ start:end ] ):
				# ~ doc 	= self.blueprint.copy()
				# ~ doc.update( item )
				# ~ results.append( doc )
			# ~ return results
		
		if isinstance( key, ObjectId ):
			result = self.database[ self.name ].find_one( { "_id" : key } )
			return self.create_document( result )
		else:
			raise ValueError( "No item found with _id: " + str( key ) )


	def __setitem__( self, key, document:Document ):
		if isinstance( key, ObjectId ):
			if key == Blueprint.root_id:
				raise ValueError( "Changing document with __root_id will change blueprint, use update_blueprint instead!" )
			else:
				if self.database[ self.name ].find_one( { "_id" : key } ):
					try:
						document.set_all_values()
						self.database[ self.name ].replace_one(  { "_id" : key }, document )
						register_database_change( self.name )
					except Exception as e:
						raise ValueError( str( e ) )
				else:

					raise ValueError( "No item found with _id: " + str( key ) )
		else:
			raise ValueError("Key needs be an ObjectdId and not a " + type( key ) )


	def __iter__( self ):
		#
		# if host is server self is local ignore 
		# of iets anders verzinnen dat je ook alle 
		# 
		kwargs = { "_id" :{ "$ne": Blueprint.root_id } }
		if self.is_local():
			kwargs[ "_local" ] = Userhost().value

		for item in self.find():
			yield self.create_document( item )


	def find( self, **kwargs ):
		if self.is_local():
			kwargs[ "_local" ] = Userhost().value

		kwargs.update( { "_id" :{ "$ne": Blueprint.root_id } } )
		results = [ self.create_document( item ) for item in self.database[ self.name ].find( kwargs, sort = [ self._sort ], limit = self._limit ) ]
		return results


	def find_in_cache( self, **kwargs ):
		if self.is_local():
			kwargs[ "_local" ] = Userhost().value
		kwargs.update( { "_id" :{ "$ne": Blueprint.root_id } } )
		return [ item["_saved"] for item in self.database[ self.name ].find( kwargs, sort = [ self._sort ], limit = self._limit ) ]


	def find_one( self, **kwargs ):
		if self.is_local():
			kwargs[ "_local" ] = Userhost().value

		kwargs.update( { "_id" :{ "$ne":  Blueprint.root_id } } )
		result = self.database[ self.name ].find_one( kwargs )

		if result == None:
			doc = None
		else:
			doc = self.create_document( result )

		return doc


#-----------------------------------------------------------------------
#
# datatype classes
# _function needs to be renamed to "on_click", "click" or something a-like
#
#-----------------------------------------------------------------------
class Datatype( dict ):
	"""
	change html to innertext
	"""
	def __init__( self ):
		dict.__init__( self, {	"_raw_value"	: None,
								"_locked" 		: False,
								"_hidden" 		: False,
								"_unique" 		: False,
								"_function"		: None,
								"_class"		: self.__class__.__name__,
								"_module"		: self.__class__.__module__,
								"_collection"	: None,
								"_innertext"	: None,
								"_href"			: None,
								} )
		self.value_setter()


	def __getattribute__( self, attr ):
		if attr == "value":
			return self.decoder( self.raw_value )
		
		elif attr in [ "function", "innertext" ]:
			value =  dict.__getitem__( self, f"_{attr}" )
			if value != None:
				return getattr( sys.modules[ value[ "_func_module" ] ], value[ "_func_name" ] )
			else:
				return value
		
		elif f"_{attr}" in dict.keys( self ):
				return dict.__getitem__( self, f"_{attr}" )

		else:
			return dict.__getattribute__( self, attr )


	def __setattr__( self, attr, value ):
		if attr == "value":
			dict.__setitem__( self, "_raw_value", self.encoder( value ) ) 
		elif attr in [ "function", "innertext" ]:
			result = value
			if attr == "innertext":
				if callable( value ):
					self._func_parser( value )
					# ~ result = value
				elif type( value ).__name__ in dir(builtins):
					def return_static_variable():
						return value
					result = return_static_variable
				else:
					raise ValueError( f"Cannot set function to a type of {type( value ).__name__}" )
		# ~ else:
			# ~ raise ValueError( f"Cannot set function to a type of {type(func).__name__}" )

			function = { "_func_module" : result.__module__, "_func_name" : result.__name__ }
			dict.__setitem__( self, f"_{attr}", function )
			if attr == "function":
				function[ "_href" ] = create_route_for_click( self )


		elif f"_{attr}" in dict.keys( self ):
			return dict.__setitem__( self, f"_{attr}", value )

		else:
			raise ValueError( f"{__class__.__name__} has no attribute {attr}" )


	def encoder( self, value ):
		return value


	def decoder( self, value ):
		return value


	def value_setter( self ):
		pass


	def _func_parser( self, func ):
		 pass


	def as_html( self, doc:Document ):
		if self.innertext == None:
			return self.value
		else:
			return self.innertext( doc )



#-----------------------------------------------------------------------
class Dictionary( Datatype ):
	def __init__( self, **kwargs ):
		Datatype.__init__( self, **kwargs )
		self.value = {}


#-----------------------------------------------------------------------
class List( Datatype ):
	def __init__( self, **kwargs ):
		Datatype.__init__( self, **kwargs )
		self.value = []


#-----------------------------------------------------------------------
class Integer( Datatype ):
	def __init__( self, **kwargs ):
		Datatype.__init__( self, **kwargs )
		self.value = 0


#-----------------------------------------------------------------------
class Boolean( Datatype ):
	def __init__( self, **kwargs ):
		Datatype.__init__( self, **kwargs )
		self.value = False


#-----------------------------------------------------------------------
class Float( Datatype ):
	def __init__( self, **kwargs ):
		Datatype.__init__( self, **kwargs )
		self.value = 0.0

#-----------------------------------------------------------------------
class String(Datatype):
	def __init__( self, **kwargs ):
		Datatype.__init__( self, **kwargs )


#-----------------------------------------------------------------------
class Null( Datatype ):
	def __init__( self, **kwargs ):
		Datatype.__init__( self, **kwargs )

	def value_setter( self ):
		 self.value = None


#-----------------------------------------------------------------------
class Date( String ):
	def __init__( self, **kwargs ):
		String.__init__( self, **kwargs )


	def value_setter( self ):
		now = datetime.now()
		self.value = now.strftime( "%d/%m/%Y %H:%M:%S" )



#-----------------------------------------------------------------------
class Path( String ):
	# use pathlib instead of os.path!
	def __init__( self, **kwargs ):
		String.__init__( self, **kwargs )


	def encode( self, value ):
		return value


	def decode( self, value ):
		# ~ print( node.__name__ )
		return value

	@property
	def exists( self ):
		return os.path.exists( self.value )


	@property
	def is_file( self ):
		return os.path.isfile( self.value )


	@property
	def is_dir( self ):
		return os.path.isdir( self.value )


	@property
	def absolute_path( self ):
		return "abs_path for: %s" % self.value


	@property
	def relative_path( self ):
		return "rel_path for: %s" % self.value


	@property
	def base_name( self ):
		return "base_name for: %s" % self.value


	@property
	def original_path( self ):
		return self._orig_value


#-----------------------------------------------------------------------
class Directory( Path ):
	def __init__( self, **kwargs ):
		Path.__init__( self, **kwargs )


class File( Path ):
	def __init__( self, **kwargs ):
		Path.__init__( self, **kwargs )

	@property
	def extension( self ):
		return os.path.splitext( self.value )[-1]


#-----------------------------------------------------------------------
class Userhost( Datatype ):
	def __init__( self, **kwargs ):
		Datatype.__init__( self, **kwargs )
		# ~ self.value = [ config.USER, config.HOST ]

	
	@property
	def user( self ):
		return self.value[1]


	@property
	def host( self ):
		return self.value[0]


	def value_setter( self ):
		self.value = [ config.HOST, config.USER ]


#-----------------------------------------------------------------------
class User( String ):
	def __init__( self, **kwargs ):
		String.__init__( self, **kwargs )

	def value_setter( self ):
		self.value = config.USER


class Host( String ):
	def __init__( self, **kwargs ):
		String.__init__( self, **kwargs )


	def value_setter( self ):
		self.value = config.HOST


#-----------------------------------------------------------------------
#
# helper functions
#
#-----------------------------------------------------------------------
def database_is_running():
	connection = pymongo.MongoClient( "%s" % config.DATABASE_IP, config.DATABASE_PORT, serverSelectionTimeoutMS = 2000 )
	try:
		return connection.server_info() # will throw an exception

	except Exception as e:
		return False


def connect_to_database( name = None ):
	if name == None:
		name = config.GESTION_DB
	connection = pymongo.MongoClient( "%s" % config.DATABASE_IP, config.DATABASE_PORT )
	return connection[ name ]


def list_databases():
	connection = pymongo.MongoClient( "%s" % config.DATABASE_IP, config.DATABASE_PORT )
	return [ db["name"] for db in connection.list_databases() if db["name"] not in [ "admin", "config", "local" ] ]


def list_collections( ):
	database = connect_to_database( )
	return database.list_collection_names()


def collection_exists( name:str ):
	return name in list( list_collections() )


def create_collection( name:str, blueprint:Blueprint ):
	coll = Collection( name )
	coll.blueprint = blueprint
	if not coll.exists: 
		coll.create()
		
	return coll


def get_collection( name:str ):
	if name in list_collections( ):
		coll = Collection( name )
		return coll
	else:
		raise Exception( "Collection %s does not exist in database %s" % ( name, config.GESTION_DB ) )


#-----------------------------------------------------------------------
def init():
	if database_is_running():
		if config.NODE_TYPE == "master":
			if config.GESTION_DB not in list_databases():
				database = connect_to_database()
	else:
		raise Exception( "Cannot connect to database on: %s" % config.DATABASE_IP )


#-----------------------------------------------------------------------
def create_route_for_click( datatype, check_exists = False ):
	# ~ print ( "create_route_for_click" )
	# ~ print( datatype )
	from modules import web
	exists 	= False
	func	= datatype.function
	# ~ print( func )
	# ~ func( "Arno" )
	# ~ exit()
	url		= f"{func.__module__}/{func.__name__}"
	# ~ print( f"generated url is {url}" )
	
	if web.server != None:
		# ~ print( "web.server != None" )
		if check_exists:
			if f"/{url}" in [ rule.rule for rule in web.server.url_map.iter_rules() ]:
				exists = True
		
		if not exists:
			# ~ print( f" {url} doesnt exist" )
			simpleclick 				= web.SimpleClick( url )
			simpleclick.needs_document 	= True
			simpleclick.function 		= func
			web.create_route( simpleclick )
	
	return url


#-----------------------------------------------------------------------
#
#	create some database use by the database module
#
#-----------------------------------------------------------------------
if database_is_running():
	changes 				= Blueprint()
	changes[ "name" ] 		= String()
	changes[ "userhost" ]	= List()
	changes_collection 		= create_collection( "_database_changes", changes )
# ~ else:
	# ~ raise Exception( "Can't connect to the database" )




#-----------------------------------------------------------------------
def check_for_changes( collection ):
	init()
	results				= changes_collection.find( name = collection, userhost = Userhost().value )
	has_changed 		= False
	
	if len( results ) > 0:
		has_changed = True
		for item in results:
			changes_collection.delete_one( item._id )

	return has_changed


#-----------------------------------------------------------------------
def register_database_change( collection ):
	init()
	if collection != changes_collection.name:
		# ~ changes_collection = get_collection( "_database_changes" )
		doc	 = changes_collection.create_document()
		doc[ "name" ] 		= collection
		doc[ "userhost" ]	= Userhost().value
		changes_collection.append( doc )



#-----------------------------------------------------------------------
if __name__ ==  "__main__":
	pass




import setuptools
# ~ https://python-packaging.readthedocs.io/en/latest/non-code-files.html

#-----------------------------------------------------------------------
with open("README.md", "r") as fh:
	long_description = fh.read()


#-----------------------------------------------------------------------
setuptools.setup(
	name							= "gestion",
	version							= 0.58,
	url								= "https://github.com/hurgar-nl/gestion.git",
	author							= "Arno de Grijs",
	author_email					= "info@hurgar.nl",
	description						= "Simple project automation manager",
	long_description 				= long_description,
	long_description_content_type 	= "text/markdown",
	packages 						= [ "gestion", "gestion.modules", "gestion.addons" ],
	install_requires 				= [ "click==8.0.1", "Flask==2.0.1", "itsdangerous==2.0.1", "Jinja2==3.0.1", "MarkupSafe==2.0.1", "psutil==5.8.0", "pymongo==3.12.0", "Werkzeug==2.0.1" ],
	include_package_data			= True,
	classifiers						=[
										"Programming Language :: Python :: 3",
										"License :: OSI Approved :: MIT License",
										"Operating System :: OS Independent",
									 ],
 )



#-----------------------------------------------------------------------
import time
import subprocess
import os
import shutil
import re
# ~ import moho 

#-----------------------------------------------------------------------
# ~ import config
# ~ config.load()

#-----------------------------------------------------------------------
# ~ from modules import addon
# ~ from modules import gestion


#-----------------------------------------------------------------------
def removeVersion(path):
	pattern= r"_[vV]\d*_"
	
	base, head = os.path.split(path)
	filename ,extension= os.path.splitext(head)

	results=re.search( pattern, filename )
	
	if results:
		match = results.group()
		result = filename.split(match)[0]
	else:
		result = filename
		
	return result


#-----------------------------------------------------------------------
def Test(path, end, start):
	print("*"*16)
	print(path)
	print(end)
	print(start)
	print("*"*16)


#-----------------------------------------------------------------------
# ~ test=addon.Addon()
# ~ test.set_function(Test)


#-----------------------------------------------------------------------
def GetShotsFolder(path):
	found=None
	while path:
		path=os.path.split(path)[0]
		if "shots" == os.path.split( path )[1].lower():
			found=path
			break
	
	return found
	
#-----------------------------------------------------------------------
def replace_server_path(path):
	results = []
	previous_root = None
	root = path

	while root and root != previous_root:
		previous_root = root
		root, tip = os.path.split(root)
		results.insert(0, tip)

	if "Club Baboo" in results:
		index = results.index("Club Baboo")
		new_path = os.path.join( "Y:", *results[index:])
	else:
		new_path = path
	
	return new_path

#-----------------------------------------------------------------------# command line opties voor moho
# http://www.lostmarble.com/forum/viewtopic.php?t=1318
#-----------------------------------------------------------------------
def replace_images_in_file(path):
	m = moho.MOHO(path)
	new_path = path
	is_changed=False

	new_base = None
	
	for layer in m.get_imagelayers(recursive=True):
		oldpath = layer["image_path"]
		newpath = layer["image_path"]
		
		
		if "megagamma" in oldpath.lower():
			splitted = oldpath.lower().split("megagamma")
			start = len(oldpath)-len(splitted[-1])
			newpath = "Y:"+  oldpath[start:]
			
		
		elif "meggamma" in oldpath.lower():
			splitted = oldpath.lower().split("meggamma")
			start = len(oldpath)-len(splitted[-1])
			newpath = "Y:"+  oldpath[start:]
		
		elif "omicron" in oldpath.lower():
			splitted = oldpath.lower().split("omicron")
			start = len(oldpath)-len(splitted[-1])
			newpath = "Z:"+  oldpath[start:]
		

		if oldpath != newpath:
			is_changed = True
			layer["image_path"] = newpath
			layer["image_fileref"]["path"]= newpath
	
	# save to a local folder to keep the server clean
	_dir, _file = os.path.split(path)
	base, ext = os.path.splitext(_file)
	new_file_name = base + "_RENDER" + ext
	
	render_folder = os.path.join(config.HOME, "gestion", "renderfiles")
	if not os.path.exists(render_folder):
		os.makedirs(render_folder)
	
	new_path = os.path.join(render_folder, new_file_name)

	m.save(new_path)
	return new_path
		
	

#-----------------------------------------------------------------------
def Render(file, start=None, end=None, project=None, replace_paths=True):
	from modules import client
	local_file=client.map_to_client(file)
	renderer=client.get_renderer("moho")
	
	if os.path.exists(local_file):
		if renderer!=None:
			print( "Found Moho executable: ", renderer)
			print( "Going to render file: ", local_file)		

			# check for missing image files within moho file
			new_file=replace_images_in_file(local_file)

			cli=[renderer]
			cli.append("-r")
			cli.append(new_file)
			cli.append("-f")
			cli.append("PNG")
			
			if start!=None:
				cli.append("-start")
				cli.append(str(start))
			if end!=None:
				cli.append("-end")
				cli.append(str(end))

			
			print("Project: ", project)
			
			workspace = client.get_project(project)
			print(workspace)
			local_root="Y:\Club Baboo"		
			print("local_root: ", local_root)	

			outputdir=removeVersion(file)
			outputpath=os.path.normpath(os.path.join(local_root, workspace["images"], outputdir, "%s" % outputdir))
			
			if not os.path.exists(os.path.split(outputpath)[0]):
				os.makedirs(os.path.split(outputpath)[0])

			cli.append("-o")
			cli.append(outputpath)
			
			proc=subprocess.Popen( cli, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True )
			print(proc.stderr.readlines())
			if proc.returncode != 0:
				print( "Error rendering %s " % file )
				return False
			else:
				return True

		# ~ else:
			# ~ print("No moho renderer found, skipping")
			# ~ raise ValueError( "No moho renderer found, skipping"  )
		
		#time.sleep(5)

		print("*"*16)
		return True

#-----------------------------------------------------------------------
# ~ render=addon.Addon()
# ~ render.set_function(Render)

# ~ print(replace_server_path("/volumes/megagamma/Club Baboo/SHANGHAI_ACROSSMEDIA_ART/Marketing Material/20210601_PromoArt/"))
# ~ print( replace_images_in_file("/volumes/megagamma/Club Baboo/SHOTS/MagicShow/MS16_Rubberboot/Shots/061_VeloSprong/MS16_061_VeloSprong_v016_L.moho"))






#-----------------------------------------------------------------------
# ~ import argparse
import importlib


#-----------------------------------------------------------------------
available_commands = {}


#-----------------------------------------------------------------------
def add_command( function ):
	# check doen of function ook echt een function is
	# en iets doen met de arguments, dus kijken welke arguments in 
	# de functie gevraagd worden en deze niet invoeren als dat niet nodig is
	#~ misschien de naam van de functie module.name
	#
	available_commands[ function.__name__ ] = function.__module__


#-----------------------------------------------------------------------
def add_commands( *functions ):
	for function in functions:
		add_command( function )


#-----------------------------------------------------------------------
def list_commands():
	if len( available_commands ) > 0:
		print( f"Available commands on {config.HOST}" )
		print( "#"*24 )
		for key in available_commands.keys():
			print( f"{key}" )
		print( "#"*24 )
	else:
		print( f"No commands available on {config.HOST}" )

add_command( list_commands )


#-----------------------------------------------------------------------
def parse_args_and_execute( argv ):
	gestion_path 	= argv.pop( 0 )
	command 		= argv.pop( 0 ) 

	args 	= []
	kwargs 	= {}

	index = 0
	while len( [ item for item in argv if item is not None  ] ) != 0:
		if argv[ index ].startswith( "-" ):
			
			kwargs[ argv[ index ][ 1: ] ] = argv[ index+1 ]
			argv[ index ] 	 = None
			argv[ index+ 1 ] = None
			index += 2
		else:
			args.append( argv[ index ] )
			argv[ index ] = None
			index += 1
	
	module			= importlib.import_module( available_commands[ command ] )
	func			= getattr( module, command )
	
	# print( "parse_args_and_execute" )
	# print( args )
	# print( kwargs )

	return func( *args, **kwargs )




#-----------------------------------------------------------------------
import time


#-----------------------------------------------------------------------
if "." in __name__:
	from modules import config
	from modules import tasks
	from modules import database
	from modules import node
	from modules import web
	from modules import cli
else:
	import config
	import tasks
	import database
	import node
	import web
	import cli


#-----------------------------------------------------------------------
import addons


#-----------------------------------------------------------------------
#
# create queue collection to add to database
# and create functions for datatypes in database
#
#-----------------------------------------------------------------------
def change_status( doc ):
	print( "change_status {doc} " )
	current_status 	= doc[ "status" ]
	new_status 		= doc[ "status" ]

	if current_status == "processing":
		new_status = "kill"
	
	elif current_status in  [ "killed", "done", "error", "ignore" ]:
		new_status = "waiting"
		
	elif current_status in  [ "waiting" ]:
		new_status = "ignore"
	
	if doc[ "status" ] != new_status:
		# ~ coll			= database.get_collection( "queue" )
		doc[ "status" ] = new_status
		queue_collection[ doc._id ] = doc



queue					= database.Blueprint()

status		 			= database.String() #dictionairy, per user status?
status.value			= "waiting"
status.function			= change_status
queue[ "status" ]		= status

queue[ "addon" ]		= database.String()

queue[ "arguments" ] 	= database.Dictionary()

project 				= database.String()
project.value 			= "default"
queue[ "project" ]		= project

userhost				= database.Userhost()
userhost.hidden			= True
queue[ "userhost" ]		= userhost

queue[ "date" ]			= database.Date()


#-----------------------------------------------------------------------
#
# create log collection to add to database
# and create functions for datatypes in database
#
#-----------------------------------------------------------------------
def as_html( doc ):
	list_length = len( doc["log"] ) 
	if list_length > 1:
		log = "items"
	else:
		log = "item"	

	return f"{list_length} {log}"


log						= database.List()
log.innertext			= as_html
queue[ "log" ]			= log
queue_collection 		= database.create_collection( "queue", queue )


#-----------------------------------------------------------------------
def loop_queue():
	# error checking of tasks in task class
	while True:
		if config.STATUS == True:
			# ~ conn = database.get_collection( "queue" )
			try:
				doc = queue_collection.find_one( status = "waiting" )
					
				if doc:
					doc[ "status" ] 			= "processing"
					queue_collection[ doc._id ]	= doc

					try:
						results 		= process_queued_item( doc )
						log_id			= node.write_to_log( f"Loop_queue: {results}" )
						doc[ "log" ].insert( 0, log_id )

						
					except Exception as e:
						log_id			= node.write_to_log( f"Loop_queue: {str(e)}" )
						doc[ "log" ].insert( 0, log_id )
						doc[ "status" ] 	= "error"
					queue_collection[ doc._id ] = doc
			
			except Exception as e:
				error = str( e )
				print( error )
				node.write_to_log( f"Loop_queue: {error}" )

		else:
			pass
			print( f"Host: {config.HOST} not available for queue." )

		time.sleep( 5 )


#-----------------------------------------------------------------------
def process_queued_item( queued_item ):
	name 	= queued_item[ "addon" ]
	_id		= queued_item._id

	if name in addons.available_addons():
		addon	= addons.LOADED[ name ]
		kwargs	= queued_item[ "arguments" ]
		results = addon.execute( **kwargs )
		
		try:
			while addon.state:
				doc = queue_collection[ _id ]
				if queue_collection[ _id ][ "status" ] == "kill":
					addon.kill()
					results 		= addon.output
					doc[ "status" ] = "killed" 
					queue_collection[ _id ] = doc
				else:
					time.sleep( 1 )
			
			if doc[ "status" ] != "killed":
				doc[ "status" ] = "done"
			queue_collection[ _id ] = doc
			
			return results
		
		except Exception as exc:
			raise( f"Error occured in processing of queued item" )

	

	# ~ else:
		# ~ raise NameError( f"Addon {name} not found, not processing" )
	
	# ~ if addon.output:
		# ~ return addon.output
	# ~ elif addon.error:
		# ~ print( "addon.error" )
		# ~ raise Exception( str( addon.error ) )
	# ~ else:
		# ~ raise( "Cannot get results from addon, has it been killed" )


#-----------------------------------------------------------------------
loop_queue_task		 = tasks.Task( loop_queue )
loop_queue_task.wait = False


#-----------------------------------------------------------------------
def add_to_queue( addon, project = None, **kwargs ):
	print( f"add_to_queue( {addon}, {project}, {kwargs} )" )
	# ~ builtin a check if file and addon exist
	if project is None:
		project = config.PROJECT

	try:
		if addon in addons.available_addons():
			#
			# need a check if all arguments needed for addon are being passed 
			#
			# ~ coll 						= database.get_collection( "queue" )
			doc 						= queue_collection.create_document()
			doc[ "addon" ]				= addon
			doc[ "project" ]		  	= project
			doc[ "arguments" ]		 	= kwargs
			doc[ "log" ].append( node.write_to_log( f"Added {addon} to queue by {config.USER}" ) )
			# ~ print( doc )
			queue_collection.append( doc )


		else:
			raise NameError( f"Addon {addon} not found, ignoring request and not adding to queue" )

	except Exception as e:
		print( str( e ) )
		node.write_to_log( str( e ) )

cli.add_command( add_to_queue )


#-----------------------------------------------------------------------
#
# create queue webpage
#
#-----------------------------------------------------------------------
queue_template 	= web.CollectionTemplate( queue_collection )
queue_page 		= web.Page( "queue", template = queue_template )




